module: malbolge
synopsis: A program language from hell.
author: iwata <iratqq@gmail.com>
copyright: see below

//
// Copyright (c) iwata <iratqq@gmail.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//


define constant xlat1 = #[#x2b, #x62, #x28, #x32, #x39, #x65, #x2a, #x6a, #x31, #x56,
                          #x4d, #x45, #x4b, #x4c, #x79, #x43, #x7d, #x29, #x38, #x26,
                          #x6d, #x23, #x7e, #x57, #x3e, #x71, #x78, #x64, #x52, #x70,
                          #x30, #x77, #x6b, #x72, #x55, #x6f, #x5b, #x44, #x37, #x2c,
                          #x58, #x54, #x63, #x41, #x22, #x6c, #x49, #x2e, #x76, #x25,
                          #x7b, #x67, #x4a, #x68, #x34, #x47, #x5c, #x2d, #x3d, #x4f,
                          #x40, #x35, #x60, #x5f, #x33, #x69, #x3c, #x3f, #x5a, #x27,
                          #x3b, #x46, #x4e, #x51, #x75, #x59, #x5d, #x73, #x7a, #x66,
                          #x24, #x21, #x42, #x53, #x2f, #x7c, #x74, #x3a, #x50, #x6e,
                          #x36, #x5e, #x48, #x61];
define constant xlat2 = #[#x2b, #x62, #x28, #x32, #x39, #x65, #x2a, #x6a, #x31, #x56,
                          #x4d, #x45, #x4b, #x4c, #x79, #x43, #x7d, #x29, #x38, #x26,
                          #x6d, #x23, #x7e, #x57, #x3e, #x71, #x78, #x64, #x52, #x70,
                          #x30, #x77, #x6b, #x72, #x55, #x6f, #x5b, #x44, #x37, #x2c,
                          #x58, #x54, #x63, #x41, #x22, #x6c, #x49, #x2e, #x76, #x25,
                          #x7b, #x67, #x4a, #x68, #x34, #x47, #x5c, #x2d, #x3d, #x4f,
                          #x40, #x35, #x60, #x5f, #x33, #x69, #x3c, #x3f, #x5a, #x27,
                          #x3b, #x46, #x4e, #x51, #x75, #x59, #x5d, #x73, #x7a, #x66,
                          #x24, #x21, #x42, #x53, #x2f, #x7c, #x74, #x3a, #x50, #x6e,
                          #x36, #x5e, #x48, #x61];

define constant opcodes = #[#x6a, #x69, #x2a, #x70, #x3c, #x2f, #x76, #x6f];

define constant p9 = #[1.0, 9.0, 81.0, 729.0, 6561.0];
define constant o  = #[#[ 4, 3, 3, 1, 0, 0, 1, 0, 0 ],
                       #[ 4, 3, 5, 1, 0, 2, 1, 0, 2 ],
                       #[ 5, 5, 4, 2, 2, 1, 2, 2, 1 ],
                       #[ 4, 3, 3, 1, 0, 0, 7, 6, 6 ],
                       #[ 4, 3, 5, 1, 0, 2, 7, 6, 8 ],
                       #[ 5, 5, 4, 2, 2, 1, 8, 8, 7 ],
                       #[ 7, 6, 6, 7, 6, 6, 4, 3, 3 ],
                       #[ 7, 6, 8, 7, 6, 8, 4, 3, 5 ],
                       #[ 8, 8, 7, 8, 8, 7, 5, 5, 4 ]];

define function malbolge-hash-op(x :: <integer>, y :: <integer>)
  let i = 0;
  for (j from 0 to 4 by 1)
    i := i + (o[truncate(modulo(as(<float>, y) / p9[j], 9))]
                [truncate(modulo(as(<float>, x) / p9[j], 9))]
                * truncate(p9[j]));
  end for;
  i;
end function malbolge-hash-op;

define class <malbolge> (<object>)
  slot reg-a :: <integer>,
    init-value: 0;
  slot reg-c :: <integer>,
    init-value: 0;
  slot reg-d :: <integer>,
    init-value: 0;
  slot mem :: <vector> = #[];
end class <malbolge>;

define method load(m :: <malbolge>, str :: <string>)
  local method malbolge-validate(str :: <string>)
          local method isspace(c :: <character>)
                  (c = ' ') | (c = '\f') | (c = '\n') | (c = '\r') | (c = '\t'); // | (c = '\v')
          end method isspace;
          let ret = foldl(method (c, acc)
                            let count = head(acc);
                            let mem = tail(acc);
                            if (isspace(c))
                              acc;
                            else
                              let int = as(<integer>, c);
                              let op = xlat1[modulo(int - 33 + count, 94)];
                              if (~member?(op, opcodes))
                                format-out("invalid expression %c, ignored.\n", c);
                                acc;
                              else
                                pair(count + 1, concatenate(mem, list(int)));
                              end if;
                            end if;
                          end method,
                          pair(0, #()),
                          as(<list>, str));
          tail(ret);
  end method malbolge-validate;
  local method malbolge-hash-mem(len :: <integer>)
          for (i from len to 59049 - 1)
            m.mem[i] := malbolge-hash-op(m.mem[i - 1], m.mem[i - 2]);
          end for;
          m.mem;
  end method malbolge-hash-mem;
  local method malbolge-compile(str :: <string>)
          let prog = malbolge-validate(str);
          m.mem := as(<vector>, concatenate(prog, make(<list>, size: 59049 - size(prog))));
          malbolge-hash-mem(size(prog));
  end method malbolge-compile;
  m.mem := malbolge-compile(str);
end method load;

define method exec(m :: <malbolge>)
  local method eval-loop()
          block (quit)
            if (m.mem[m.reg-c] < 33 | m.mem[m.reg-c] > 126)
              quit(#t);
            else
              let op = xlat1[modulo(m.mem[m.reg-c] - 33 + m.reg-c, 94)];
              case
                op = opcodes[0] =>
                  m.reg-d := m.mem[m.reg-d];
                op = opcodes[1] =>
                  m.reg-c := m.mem[m.reg-d];
                op = opcodes[2] =>
                  m.reg-a := m.mem[m.reg-d] := m.mem[m.reg-d] / 3 + modulo(m.mem[m.reg-d], 3) * 19683;
                op = opcodes[3] =>
                  m.reg-a := m.mem[m.reg-d] := malbolge-hash-op(m.reg-a, m.mem[m.reg-d]);
                op = opcodes[4] =>
                  format-out("%c", as(<character>, logand(m.reg-a, #xff)));
                op = opcodes[5] =>
                  let line = read-line(*standard-input*);
                  if (line.empty?)
                    format-out("empty?");
                  else
                    m.reg-a := logand(as(<integer>, head(as(<list>, copy-sequence(line, start: 0, end: 1)))),
                                #xff);
                  end if;
                op = opcodes[6] => quit(#t);
              end case;
              m.mem[m.reg-c] = xlat2[m.mem[m.reg-c] - 33];
              if (m.reg-c = 59048)
                m.reg-c := 0;
              else
                m.reg-c := m.reg-c + 1;
              end if;
              if (m.reg-d = 59048)
                m.reg-d := 0;
              else
                m.reg-d := m.reg-d + 1
              end if;
              eval-loop();
            end if;
          end block;
  end method eval-loop;
  eval-loop();
end method exec;

define method usage()
  format-out("usage: malbolge file\n");
end method usage;

define function main(name, arguments)
  if (size(arguments) = 1)
    with-open-file(s = arguments[0])
      let m = make(<malbolge>);
      load(m, read-to-end(s));
      exec(m);
    end with-open-file;
  else
    usage();
  end if;
  exit-application(0);
end function main;

// Invoke our main() function.
main(application-name(), application-arguments());
